/*
 * QtCurve GTK engine
 *
 * This is a hacked/modified version of the B***Curve GTK engine. Modified to use
 * ~/.qt/qtrc to get colour information to recolour widgets.
 *
 * Changes are (C) Craig Drummond 2003 - Craig.Drummond@lycos.co.uk
 *
 */

#include "qt_settings.h"
#include <gmodule.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtkenums.h>

static int strcmp_i(const char *s1, const char *s2)
{
    char c1,
         c2;

    for(;;)
    {
        c1=*s1++;
        c2=*s2++;
        if(!c1 || !c2)
            break;
        if(isupper(c1))
            c1=tolower (c1);
        if(isupper(c2))
            c2=tolower (c2);
        if(c1!=c2)
            break;
    }
    return (int)c2-(int)c1;
}

struct qt_data qt_settings;

static int ctoh(char ch)
{
    return (ch>='0' && ch<='9') ? ch-'0' :
           (ch>='a' && ch<='f') ? 10+(ch-'a') :
           (ch>='A' && ch<='F') ? 10+(ch-'A') :
           0;
}

#define ATOH(str) ((ctoh(*str)<<4)+ctoh(*(str+1)))

static void set_rgb(GdkColor *col, char *str)
{
    if(str && strlen(str)>5)
    {
        col->red=ATOH(str)<<8;
        col->green=ATOH(&str[2])<<8;
        col->blue=ATOH(&str[4])<<8;
        col->pixel=0;
    }
    else
        col->red=col->green=col->blue=col->pixel=0;
}

#define MAX_INPUT_LINE_LEN 1024
#define MAX_FILENAME_LEN   256

enum
{
    SECT_NONE,
    SECT_PALETTE,
    SECT_GENERAL
};

/*
  Qt uses the following predefined weights: 
    Light    = 25,
    Normal   = 50,
    DemiBold = 63,
    Bold     = 75,
    Black    = 87

  ...need to categorize mid way, i.e. 0->37 = light, 38->56 = normal, etc...
*/

enum
{
    WEIGHT_NORMAL   = 38,   /* Less than this = light */
    WEIGHT_DEMIBOLD = 57,
    WEIGHT_BOLD     = 69,
    WEIGHT_BLACK    = 81
};

static const char * weight_str(int w)
{
    if(w<WEIGHT_NORMAL)
        return "light";
    else if(w<WEIGHT_DEMIBOLD)
#ifdef GTK2_VER
        return "";
#else
        return "medium";
#endif
    else if(w<WEIGHT_BOLD)
        return "demibold";
    else if(w<WEIGHT_BLACK)
        return "bold";
    else
        return "black";
}

#ifdef GTK2_VER
static const char * italic_str(int i)
{
    return i ? "Italic" : "";
}
#endif

enum
{
    RD_PALETTE = 0x1,
    RD_FONT    = 0x2
};

#ifndef GTK2_VER
static const char * iso8859_1_langs[]=
{
    "af", "afrikaans", "ca", "catalan", "da", "danish", "dansk", "de", "german", "deutsch",
    "en", "english", "es", "spanish", "eu", "basque", "fi", "finnish", "fo", "faroese",
    "faeroese", "fr", "french", "ga", "irish", "gd", "scottish", "gl", "galician", "is",
    "icelandic", "it", "italian", "nl", "dutch", "no", "norwegian", "pt", "portuguese",
    "sv", "swedish",
    NULL
};

static const char * iso8859_1_codesets[]=
{
    "ISO8859-1", "ISO_8859-1", "iso88591", "88591", "88591.en", "8859", "8859.in", "ascii",
    NULL
};

static const char * iso8859_2_langs[]=
{
    "cs", "croatian", "cz", "czech", "hr", "hu", "hungarian", "pl", "polish", "ro", "romanian",
    "rumanian", "serbocroatian", "sh", "sk", "sl", "slovak", "sr", "slovene", "slovenian",
    "sq", "albanian",
    NULL
};

static const char * iso8859_2_codesets[]=
{
    "ISO8859-2", "ISO_8859-2", "iso88592", "88592",
    NULL
};

static const char * iso8859_3_langs[] =
{
    "eo", "esperanto", "mt", "maltese",
    NULL
};

static const char * iso8859_3_codesets[]=
{
    "ISO8859-3", "ISO_8859-3", "iso88593", "88593",
    NULL
};

static const char * iso8859_5_langs[] =
{
    "be", "byelorussian", "bg", "bulgarian", "mk", "macedonian", "sp", 
    "sr", "serbian",
    NULL
};

static const char * iso8859_5_codesets[]=
{
    "ISO8859-5", "ISO_8859-5", "iso88595", "88595",
    NULL
};

static const char * iso8859_6_langs[] =
{
    "ar", "arabic",
    NULL
};

static const char * iso8859_6_codesets[]=
{
    "ISO8859-6", "ISO_8859-6", "iso88596", "88596",
    NULL
};

static const char * iso8859_7_langs[] =
{
    "el", "greek",
    NULL
};

static const char * iso8859_7_codesets[]=
{
    "ISO8859-7", "ISO_8859-7", "iso88597", "88597",
    NULL
};

static const char * iso8859_8_langs[] =
{
    "hebrew", "iw", "he",
    NULL
};

static const char * iso8859_8_codesets[]=
{
    "ISO8859-8", "ISO_8859-8", "iso88598", "88598",
    NULL
};

static const char * iso8859_9_langs[] =
{
    "tr", "tr", "turkish",
    NULL
};

static const char * iso8859_9_codesets[]=
{
    "ISO8859-9", "ISO_8859-9", "iso88599", "88599",
    NULL
};

static const char * iso8859_10_langs[] =
{
    "et", "estonian", "lt", "lithuanian", "lv", "latvian",
    NULL
};

static const char * iso8859_10_codesets[]=
{
    "ISO8859-10", "ISO_8859-10", "iso885910", "885910",
    NULL
};

static const char * iso8859_15_langs[] =
{
/*
      "fr", "french", "fi", "finnish",
*/
    NULL
};

static const char * iso8859_15_codesets[]=
{
    "ISO8859-15", "ISO_8859-15", "iso885915", "885915",
    NULL
};

static const char * koi8_r_langs[] =
{
    "ru", "russian",
    NULL
};

static const char * koi8_r_codesets[]=
{
    "KOI8-R", "KOI8_R",
    NULL
};

static const char * koi8_ru_langs[] =
{
    "uk", "ukrainian",
    NULL
};

static const char * koi8_ru_codesets[]=
{
    "KOI8-RU", "KOI8_RU",
    NULL
};

static struct
{
    const char *encoding;
    const char **languages,
               **codesets;
} x11EncodingMap[] =
    {
        { "iso8859-1",  iso8859_1_langs,  iso8859_1_codesets  },
        { "iso8859-2",  iso8859_2_langs,  iso8859_2_codesets  },
        { "iso8859-3",  iso8859_3_langs,  iso8859_3_codesets  },
        { "iso8859-5",  iso8859_5_langs,  iso8859_5_codesets  },
        { "iso8859-6",  iso8859_6_langs,  iso8859_6_codesets  },
        { "iso8859-7",  iso8859_7_langs,  iso8859_7_codesets  },
        { "iso8859-8",  iso8859_8_langs,  iso8859_8_codesets  },
        { "iso8859-9",  iso8859_9_langs,  iso8859_9_codesets  },
        { "iso8859-10", iso8859_10_langs, iso8859_10_codesets },
        { "iso8859-15", iso8859_15_langs, iso8859_15_codesets },
        { "KOI8-R",     koi8_r_langs,     koi8_r_codesets     },
        { "KOI8-RU",    koi8_ru_langs,    koi8_ru_codesets    },
        { NULL,         NULL,             NULL                }
    };

static const char * getX11Encoding()
{
    const char *enc=getenv("QT_GTK_ENC");

    if(!enc)
        enc=getenv("QTPIXMAP_GTK_ENC");   /* For backwards compatibility */

    if(!enc)
    {
        const char *locale;

        locale = getenv("LC_ALL");
        if (!locale || !*locale)
        { 
            locale = getenv("LC_CTYPE");
            if (!locale || !*locale)
                locale = getenv("LANG");
        }

        if (locale && *locale)
        {
            /* The most general syntax of a locale (not all optional parts recognized by all systems) is
               language[_territory][.codeset][@modifier][+special][,[sponsor][_revision]]
               To retrieve the codeset, search the first dot. Stop searching when
               a '@' or '+' or ',' is encountered. */
            char       *buf = (char*) malloc(strlen(locale)+1);
            const char *codeset = NULL;
            const char *cp = locale;

            for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++)
            {
                if (*cp == '.')
                {
                    codeset = ++cp;

                    for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++)
                        ;
                    if (*cp != '\0')
                    {
                        size_t n = cp - codeset;
                        memcpy(buf, codeset, n);
                        buf[n] = '\0';
                        codeset = buf;
                    }
                    break;
                }
            }

            if (codeset)
            {
                int i;

                for(i=0; x11EncodingMap[i].encoding && !enc; ++i)
                {
                    int j;

                    for(j=0; x11EncodingMap[i].codesets[j] && !enc; ++j)
                        if(0==strcmp_i(codeset, x11EncodingMap[i].codesets[j]))
                            enc=x11EncodingMap[i].encoding;
                }

                if(!enc)  /* Then didn't recognise codeset... */
                {
                    char *dash=strchr(codeset, '-');

                    /* Check if codeset is of the form <wibble>-<wobble> */
                    if(dash && strlen(dash)>1 && NULL==strchr(dash+1, '-'))  /* if so, then assume this is an OK X11 enc... */
                        enc=codeset;
                }
            }

            if(!enc) /* No matching/recognised codeset, so try to guess based upon language */
            {
                const char *underscore = strchr(locale, '_');
                const char *lang;
                int        i;

                if (underscore)
                {
                    size_t n = underscore - locale;
                    memcpy(buf,locale,n);
                    buf[n] = '\0';
                    lang = buf;
                }
                else
                    lang = locale;

                for(i=0; x11EncodingMap[i].encoding && !enc; ++i)
                {
                    int j;

                    for(j=0; x11EncodingMap[i].languages[j] && !enc; ++j)
                        if(0==strcmp_i(lang, x11EncodingMap[i].languages[j]))
                            enc=x11EncodingMap[i].encoding;
                }
            }

            free(buf);
        }
    }

    return enc ? enc : "iso8859-1";
}

static void getX11Res(int *resX, int *resY)
{
    gint width=gdk_screen_width(),
         height=gdk_screen_height(),
         widthMM=gdk_screen_width_mm(),
         heightMM=gdk_screen_height_mm();

    *resX=(width*254+widthMM*5)/(widthMM*10);
    *resY=(height*254+heightMM*5)/(heightMM*10);
}
#endif

int read_font_replacement(const char *file, const char **font)
{
    FILE *r=fopen(file, "r");
    int  ret=0;


    if(r)
    {
        char        line[MAX_INPUT_LINE_LEN+1];
        static char font_line[MAX_INPUT_LINE_LEN+1];

        while(NULL!=fgets(line, MAX_INPUT_LINE_LEN, r))
            if(line[0] != '#')
            {
                char *of=NULL;

                memcpy(font_line, line, MAX_INPUT_LINE_LEN+1);
                of=strtok(font_line, "=");
                if(0==strcmp_i(*font, of))
                {
                    *font=strtok(NULL, "\n");
                    ret=1;
                    break;
                }
            }
        fclose(r);
    }
    return ret;
}

static int read_rc(const char *rc, int rd)
{
    char       *home=getenv("HOME");
    int        found=0,
               weight=WEIGHT_NORMAL,
               italic=0,
               fixedW=0;
    float      size=12.0;
    const char *family="helvetica",
               *foundry="adobe";
    char       line[MAX_INPUT_LINE_LEN+1],
               font_line[MAX_INPUT_LINE_LEN+1];

    if(NULL!=home)
    {
        char fname[256];
        FILE *f;

        sprintf(fname, "%s/%s", home, rc);

        f=fopen(fname, "r");

        if(f)
        {
            int section=SECT_NONE;

            while(found!=rd && NULL!=fgets(line, MAX_INPUT_LINE_LEN, f))
                if(line[0]=='[')
                {
                    if(0==memcmp(line, "[Palette]", 9))
                        section=SECT_PALETTE;
                    else if(0==memcmp(line, "[General]", 9))
                        section=SECT_GENERAL;
                    else
                        section=SECT_NONE;
                }
                else if(SECT_PALETTE==section && rd&RD_PALETTE && !(found&RD_PALETTE) && 0==memcmp(line, "active=", 7))
                {
                    int  n=-1;
                    char *l=strtok(line, "#");

                    found|=RD_PALETTE;

                    while(l)
                    {
                        if(8==strlen(l))
                            switch(n)
                            {
                                case 0:
                                    set_rgb(&qt_settings.colors[COLOR_FOREGROUND], l);
                                    break;
                                case 1:
                                    set_rgb(&qt_settings.colors[COLOR_BUTTON], l);
                                    break;
                                case 5:
                                    set_rgb(&qt_settings.colors[COLOR_MID], l);
                                    break;
                                case 6:
                                    set_rgb(&qt_settings.colors[COLOR_TEXT], l);
                                    break;
                                case 9:
                                    set_rgb(&qt_settings.colors[COLOR_BACKGROUND], l);
                                    break;
                                case 10:
                                    set_rgb(&qt_settings.colors[COLOR_WINDOW], l);
                                    break;
                                case 12:
                                    set_rgb(&qt_settings.colors[COLOR_SELECTED], l);
                                    break;
                                case 13:
                                    set_rgb(&qt_settings.colors[COLOR_TEXT_SELECTED], l);
                                    break;
                                default:
                                    break;
                            }
                        else
                            if(n>-1)
                                break;

                        n++;
                        if(n>13)
                            break;
                        l=strtok(NULL, "#");
                    }
                }
                else if (SECT_GENERAL==section && rd&RD_FONT && !(found&RD_FONT) && 0==memcmp(line, "font=", 5))
                {
                    int   n=-1,
                          rc_weight=WEIGHT_NORMAL,
                          rc_italic=0,
                          rc_fixedW=0;
                    float rc_size=12.0;
                    char  *l,
                          *rc_family=NULL,
                          *rc_foundry=NULL;

                    memcpy(font_line, line, MAX_INPUT_LINE_LEN+1);
                    l=strtok(font_line, "=");
                    found|=RD_FONT;

                    while(l)
                    {
                        switch(n)
                        {
                            case 0:  /* Family - and foundry(maybe!) (ignore X11 and XFT) */
                            {
                                char *ob=NULL,
                                     *cb=NULL;

                                if(NULL!=(ob=strchr(l, '[')) && ob!=l && NULL!=(cb=strchr(l, ']')))
                                {
                                    ob[-1]='\0';
                                    *cb='\0';
                                    rc_foundry=&(ob[1]);

                                    if(0==strcmp_i(rc_foundry, "xft") || 0==strcmp_i(rc_foundry, "x11"))
                                        rc_foundry=NULL;
                                }
                                else  /* Sometimes .kderc has "adobe-helvetica" */
                                {
                                    char *dash=NULL;

                                    if(NULL!=(dash=strchr(l, '-')))
                                    {
                                        rc_foundry=l;
                                        *dash='\0';
                                        l=++dash;
                                    }
                                }

                                rc_family=l;
                                break;
                            }
                            case 1:  /* Point size */
                                sscanf(l, "%f", &rc_size);
                                break;
                            case 4:  /* Weight */
                                sscanf(l, "%d", &rc_weight);
                                break;
                            case 5:  /* Slant */
                                sscanf(l, "%d", &rc_italic);
                                break;
                            case 8:  /* Spacing */
                                sscanf(l, "%d", &rc_fixedW);
                                break;
                            default:
                                break;
                        }

                        n++;
                        if(n>8 && NULL!=family)
                        {
                            weight=rc_weight;
                            italic=rc_italic;
                            fixedW=rc_fixedW;
                            size=rc_size;
                            family=rc_family;
                            foundry=rc_foundry;
                            break;
                        }
                        l=strtok(NULL, ",");
                    }
                }
                else if(found==rd)
                    break;

            fclose(f);
        }
    }

    if(rd&RD_PALETTE && !(found&RD_PALETTE))
    {
        qt_settings.colors[COLOR_BACKGROUND].red=255<<8;
        qt_settings.colors[COLOR_BACKGROUND].green=255<<8;
        qt_settings.colors[COLOR_BACKGROUND].blue=255<<8;
        qt_settings.colors[COLOR_BACKGROUND].pixel=0;
        qt_settings.colors[COLOR_BUTTON].red=238<<8;
        qt_settings.colors[COLOR_BUTTON].green=234<<8;
        qt_settings.colors[COLOR_BUTTON].blue=222<<8;
        qt_settings.colors[COLOR_BUTTON].pixel=0;
        qt_settings.colors[COLOR_SELECTED].red=255<<8;
        qt_settings.colors[COLOR_SELECTED].green=221<<8;
        qt_settings.colors[COLOR_SELECTED].blue=118<<8;
        qt_settings.colors[COLOR_SELECTED].pixel=0;
        qt_settings.colors[COLOR_WINDOW].red=238<<8;
        qt_settings.colors[COLOR_WINDOW].green=238<<8;
        qt_settings.colors[COLOR_WINDOW].blue=230<<8;
        qt_settings.colors[COLOR_WINDOW].pixel=0;

        qt_settings.colors[COLOR_FOREGROUND].red=0<<8;
        qt_settings.colors[COLOR_FOREGROUND].green=0<<8;
        qt_settings.colors[COLOR_FOREGROUND].blue=0<<8;
        qt_settings.colors[COLOR_FOREGROUND].pixel=0;
        qt_settings.colors[COLOR_MID].red=198<<8;
        qt_settings.colors[COLOR_MID].green=198<<8;
        qt_settings.colors[COLOR_MID].blue=191<<8;
        qt_settings.colors[COLOR_MID].pixel=0;
        qt_settings.colors[COLOR_TEXT].red=0<<8;
        qt_settings.colors[COLOR_TEXT].green=0<<8;
        qt_settings.colors[COLOR_TEXT].blue=0<<8;
        qt_settings.colors[COLOR_TEXT].pixel=0;
        qt_settings.colors[COLOR_TEXT_SELECTED].red=0<<8;
        qt_settings.colors[COLOR_TEXT_SELECTED].green=0<<8;
        qt_settings.colors[COLOR_TEXT_SELECTED].blue=0<<8;
        qt_settings.colors[COLOR_TEXT_SELECTED].pixel=0;
    }

    if(rd&RD_FONT)  /* No need to check if read in */
    {
#ifdef GTK2_VER
        qt_settings.font=(char *)malloc(
            strlen(family) + 1 +
            strlen(weight_str(weight)) + 1 +
            strlen(italic_str(weight)) + 1 +
            20+  /* point size */ +1);

        sprintf(qt_settings.font, "%s %s %s %d",
                family,
                weight_str(weight),
                italic_str(italic),
                (int)size);
#else
        char       *format=(char *)"-%s-%s-%s-%s-*--*-%d-%d-%d-%s-*-%s";
        const char *encoding=getX11Encoding();
        int        resX,
                   resY;
        const char *global_name="/etc/X11/qt_gtk_fnt2fntrc";

        getX11Res(&resX, &resY);

        if(NULL!=home)
        {
            char home_name[MAX_FILENAME_LEN+1];

            snprintf(home_name, MAX_FILENAME_LEN, "%s/%s", home, "/.qt/gtk_fnt2fntrc");

            if(!read_font_replacement(home_name, &family))
                read_font_replacement(global_name, &family);
        }
        else
            read_font_replacement(global_name, &family);

        qt_settings.font=(char *)malloc(
            (foundry ? strlen(foundry) : 1) +
            strlen(family) +
            strlen(weight_str(weight))+
            1+   /* Italic */
            20+  /* point size */
            1+   /* fixed width */
            20+  /* Res X */
            20+  /* Res Y */
            strlen(format) +
            strlen(encoding));

        sprintf(qt_settings.font, format,
                foundry ? foundry : "*",
                family,
                weight_str(weight),
                italic ? "i" : "r",
                (int)(size*10),
                resX,
                resY,
                fixedW ? "m" : "p",
                encoding);
#endif
#if 0
printf("REQUEST FONT: %s\n", qt_settings.font);
#endif
    }

    return found;
}

static int qt_refs=0;

void qt_init()
{
    if(0==qt_refs)
    {
        qt_settings.font=NULL;

        read_rc(".qt/qtrc", RD_PALETTE|RD_FONT);
    }
    qt_refs++;
}

void qt_exit()
{
    qt_refs--;

    if(0==qt_refs)
    {
        if(qt_settings.font)
            free(qt_settings.font);
        qt_settings.font=NULL;
    }
}

#define SET_COLOUR(st, rc, itm, ITEM, state, QTP_COL) \
    st->itm[state]=rc->color_flags[state]&ITEM ? rc->itm[state] : qt_settings.colors[QTP_COL];

static void mod_color(GdkColor *col, int mod)
{
    col->red+=mod;
    col->green+=mod;
    col->blue+=mod;
}

void qt_set_colours(GtkStyle *style, GtkRcStyle *rc_style, int prelight_mod)
{
    SET_COLOUR(style, rc_style, bg, GTK_RC_BG, GTK_STATE_NORMAL, COLOR_WINDOW)
    SET_COLOUR(style, rc_style, bg, GTK_RC_BG, GTK_STATE_SELECTED, COLOR_SELECTED)
    SET_COLOUR(style, rc_style, bg, GTK_RC_BG, GTK_STATE_INSENSITIVE, COLOR_WINDOW)
    SET_COLOUR(style, rc_style, bg, GTK_RC_BG, GTK_STATE_ACTIVE, COLOR_MID)
    SET_COLOUR(style, rc_style, bg, GTK_RC_BG, GTK_STATE_PRELIGHT, COLOR_WINDOW)
    mod_color(&style->bg[GTK_STATE_PRELIGHT], prelight_mod<<8);

    SET_COLOUR(style, rc_style, base, GTK_RC_BASE, GTK_STATE_NORMAL, COLOR_BACKGROUND)
    SET_COLOUR(style, rc_style, base, GTK_RC_BASE, GTK_STATE_SELECTED, COLOR_SELECTED)
    SET_COLOUR(style, rc_style, base, GTK_RC_BASE, GTK_STATE_INSENSITIVE, COLOR_WINDOW)
    SET_COLOUR(style, rc_style, base, GTK_RC_BASE, GTK_STATE_ACTIVE, COLOR_BACKGROUND)
    SET_COLOUR(style, rc_style, base, GTK_RC_BASE, GTK_STATE_PRELIGHT, COLOR_BACKGROUND)

    SET_COLOUR(style, rc_style, text, GTK_RC_TEXT, GTK_STATE_NORMAL, COLOR_TEXT)
    SET_COLOUR(style, rc_style, text, GTK_RC_TEXT, GTK_STATE_SELECTED, COLOR_TEXT_SELECTED)
    SET_COLOUR(style, rc_style, text, GTK_RC_TEXT, GTK_STATE_INSENSITIVE, COLOR_MID)
    SET_COLOUR(style, rc_style, text, GTK_RC_TEXT, GTK_STATE_ACTIVE, COLOR_TEXT)
    SET_COLOUR(style, rc_style, text, GTK_RC_TEXT, GTK_STATE_PRELIGHT, COLOR_TEXT)

    SET_COLOUR(style, rc_style, fg, GTK_RC_FG, GTK_STATE_NORMAL, COLOR_FOREGROUND)
    SET_COLOUR(style, rc_style, fg, GTK_RC_FG, GTK_STATE_SELECTED, COLOR_TEXT_SELECTED)
    SET_COLOUR(style, rc_style, fg, GTK_RC_FG, GTK_STATE_INSENSITIVE, COLOR_MID)
    SET_COLOUR(style, rc_style, fg, GTK_RC_FG, GTK_STATE_ACTIVE, COLOR_FOREGROUND)
    SET_COLOUR(style, rc_style, fg, GTK_RC_FG, GTK_STATE_PRELIGHT, COLOR_FOREGROUND)
}

void qt_set_font(GtkRcStyle *rc_style)
{
    if(qt_settings.font)
    {
#ifdef GTK2_VER
        if (rc_style->font_desc)
            pango_font_description_free (rc_style->font_desc);

        rc_style->font_desc = pango_font_description_from_string (qt_settings.font);
#else
        rc_style->font_name = g_strdup(qt_settings.font);
#endif
    }
}
